import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer_plugin/protocol/protocol_common.dart';
import 'package:analyzer_plugin/protocol/protocol_generated.dart';
import 'package:lambda_dart/src/source_value.dart';

class Checker {
  Map<AnalysisError, PrioritizedSourceChange> check(
      LibraryElement libraryElement) {
    var result = <AnalysisError, PrioritizedSourceChange>{};

    for (var compilationUnit in libraryElement.units) {
      // Don't analyze if there's no source; there's nothing to do.
      if (compilationUnit.source == null) continue;

      // Don't analyze generated source; there's nothing to do.
      if (compilationUnit.source.fullName.endsWith('.g.dart')) continue;
      if (compilationUnit.source.fullName.endsWith('.freezed.dart')) continue;

      for (var type in compilationUnit.types) {
        final sourceClass = SourceValue(type);
        final errors = sourceClass.computeErrors();

        if (errors.isNotEmpty) {
          final lineInfo = compilationUnit.lineInfo;

          final builtNode = sourceClass
              .classDeclaration.implementsClause.interfaces
              .singleWhere((typeName) => typeName.name.name == 'Built');
          final offset = builtNode.offset;
          final length = builtNode.length;
          final offsetLineLocation = lineInfo.getLocation(offset);
          final error = AnalysisError(
              AnalysisErrorSeverity.WARNING,
              AnalysisErrorType.LINT,
              Location(
                  compilationUnit.source.fullName,
                  offset,
                  length,
                  offsetLineLocation.lineNumber,
                  offsetLineLocation.columnNumber),
              'Class has SIDE-EFFECT(S): ' +
                  errors.map((error) => error.message).join(' '),
              'SIDE_EFFECT_WARNING');

          // Fix consists of all the individual fixes, sorted so they apply
          // from the end of the file backwards, so each fix does not
          // invalidate the line numbers for the following fixes.
          final edits = errors
              .where((error) => error.fix != null)
              .map((error) => SourceEdit(error.offset, error.length, error.fix))
              .toList();
          edits.sort((left, right) => right.offset.compareTo(left.offset));

          final fix = PrioritizedSourceChange(
              1000000,
              SourceChange(
                'Apply fixes for SIDE-EFFECTS.',
                edits: [
                  SourceFileEdit(
                    compilationUnit.source.fullName,
                    compilationUnit.source.modificationStamp,
                    edits: edits,
                  )
                ],
              ));
          result[error] = fix;
        }
      }
    }

    return result;
  }
}
